using GLib;
using Gtk;
using Pango;

public class Tag
{
	public TextMark begin;
	public TextMark end;
}

public class DocumentView : SourceView
{
	private Document _document;
	public Document document { 
		get {
			return _document;
		}
		set {
			_document = value;
			set_buffer (_document);
			
			TextTag tag = new TextTag ("tag");
			tag.background = "yellow";
			document.get_tag_table ().add (tag);
			
			restore_state ();
		}
	}

	private SList<Tag> tag_list;
	private bool modified;

	public DocumentView () {
		modify_font (FontDescription.from_string (SystemSettings.get_font ()));
		//GConf.Client.get_default().notify_add ("/desktop/gnome/interface/monospace_font_name", on_font_changed);
	
		tabs_width = 4;
		indent_on_tab = true;
		highlight_current_line = true;
		auto_indent = true;
		show_line_numbers = Settings.get_show_line_numbers ();
		show_margin = Settings.get_show_margin ();
		
		cut_clipboard += on_cut_clipboard;
		copy_clipboard += on_copy_clipboard;
		paste_clipboard += on_paste_clipboard;
		focus_out_event += on_focus_out;
	}
	
	public construct (Document doc = null) {
		if ( doc != null ) {
			document = doc;
		}
	}
	
	private void save_state () {
		TextIter start_iter, end_iter;
		
		if ( document.get_selection_bounds (out start_iter, out end_iter) ) {
			int tmp = TextIter.get_offset (ref end_iter);
			document.metadata.@set ("sel_offset", tmp.to_string ());
		} else {
			document.metadata.@set ("sel_offset", "-1");
		}
		int tmp = TextIter.get_offset (ref start_iter);
		document.metadata.@set ("cur_offset", tmp.to_string ());
	}
	
	private void restore_state () {
		TextIter start_iter, end_iter;
		
		int cur = document.metadata.get_int ("cur_offset");
		int sel = document.metadata.get_int ("sel_offset");
		
		document.get_iter_at_offset (ref start_iter, cur);
		if ( sel > -1 ) {
			document.get_iter_at_offset (ref end_iter, sel);
		} else {
			end_iter = start_iter;
		}
		document.select_range (ref start_iter, ref end_iter);
	}
	
	public bool find (string text, TextIter start, bool match_case = false, bool backwards = false, bool wrap = true) {
		return_if_fail (document != null);
	
		TextIter match;
		TextIter match_end;
		SourceSearchFlags search_flags;
		bool wrapped = false;
		bool found = false;
		
		if ( !match_case ) {
			search_flags |= SourceSearchFlags.CASE_INSENSITIVE;
		}
		
		TextIter limit;	
		if ( backwards )
			document.get_start_iter (out limit);
		else
			document.get_end_iter (out limit);
		
		while (true) {
			if ( backwards ) {
				found = SourceIter.backward_search (ref start, text, search_flags, out match, out match_end, ref limit);
			} else {
				found = SourceIter.forward_search (ref start, text, search_flags, out match, out match_end, ref limit);
			}
		
			if ( found ) {
				document.select_range (ref match, ref match_end);
				scroll_to_iter (ref match_end, 0, false, 0, 0);
				return true;
			} else {
				if ( !wrap || wrapped ) {
					return false;
				}
			}
			
			if ( wrap && !wrapped) {
				limit = start;
				if ( backwards )
					document.get_end_iter (out start);
				else
					document.get_start_iter (out start);
				wrapped = true;
			}
		};

		return false;
	}
	
	public void select_all () {
		return_if_fail (document != null);
		TextIter s, e;
		document.get_bounds (out s, out e);
		document.select_range (out s, out e);
	}
	
	private ref string normalize_block_clip (string clipstring) {
		string ret = clipstring;
		if ( !clipstring.has_prefix ("\n") )
				ret = "\n" + ret;
		return ret + "\n";	
	}
	
	private void save_selection () {
		TextIter start, end;
		document.get_selection_bounds (ref start, ref end);
		document.create_mark ("saved_insert", ref start, true);
		document.create_mark ("saved_selection_bound", ref end, true);
	}
	
	private void restore_selection () {
		TextIter start, end;
		document.get_iter_at_mark (ref start, document.get_mark ("saved_insert"));
		document.get_iter_at_mark (ref end, document.get_mark ("saved_selection_bound"));
		document.select_range (ref start, ref end);
	}
	
	private void select_block (ref TextIter start, ref TextIter end) {
		TextIter.set_line_index (ref start, 0);
		if ( !TextIter.ends_line (ref end) )
			TextIter.forward_to_line_end (ref end);
		
		if ( TextIter.get_line (ref start) > 0 )
			TextIter.backward_char (ref start);
		else
			TextIter.forward_char (ref end);
		
		document.select_range (ref start, ref end);
	}
	
	public void block_cut_clipboard (bool block = false) {
		return_if_fail (document != null);
	
		TextIter start, end;	
		bool has_selection = document.get_selection_bounds (out start, out end);
		
		if ( has_selection && !block ) {
			document.cut_clipboard (BaseWindow.get_clipboard (), true);
		} else {
			int line = TextIter.get_line (ref start);
			int offset = TextIter.get_line_offset (ref start);
			select_block (ref start, ref end);
			string clipstring = normalize_block_clip (document.get_text (ref start, ref end, false));
			document.delete_interactive (ref start, ref end, true);
			
			document.get_iter_at_line (ref start, line);
			if ( offset > TextIter.get_chars_in_line (ref start)-1 )
				TextIter.set_line_offset (ref start, TextIter.get_chars_in_line (ref start)-1);
			else
				TextIter.set_line_offset (ref start, offset);
			document.select_range (ref start, ref start);
			
			BaseWindow.get_clipboard ().set_text (clipstring, -1);
		}
	}
	
	public void block_copy_clipboard (bool block = false) {
		return_if_fail (document != null);
	
		TextIter start, end;
		bool has_selection = document.get_selection_bounds (out start, out end);
		
		if ( has_selection && !block ) {
			buffer.copy_clipboard (BaseWindow.get_clipboard ());
		} else {
			save_selection ();
			select_block (ref start, ref end);
			restore_selection ();
			string clipstring = normalize_block_clip (document.get_text (ref start, ref end, false));
			BaseWindow.get_clipboard ().set_text (clipstring, -1);
		}
	}
	
	public void block_paste_clipboard () {
		return_if_fail (document != null);
	
		TextIter iter, end;
		string clipstring = BaseWindow.get_clipboard ().wait_for_text ();
		
		document.get_selection_bounds (out iter, out end);
		
		if ( clipstring.has_prefix ("\n") && clipstring.has_suffix ("\n") ) {
			String str = new String (BaseWindow.get_clipboard ().wait_for_text ());
			str.erase (str.len -1, 1);
			if ( !TextIter.ends_line (ref iter) )
				TextIter.forward_to_line_end (ref iter);
			document.insert_interactive (ref iter, str.str, -1, true);
		} else {
			document.paste_clipboard (BaseWindow.get_clipboard (), ref iter, true);
		}
	}
	
	public int get_current_line () {
		return_if_fail (document != null);
	
		TextIter iter;
		document.get_iter_at_mark (ref iter, document.get_insert ());
		return TextIter.get_line (ref iter);
	}
	
	public void go_to_line (int linenum) {
		TextIter iter;
		
		if ( linenum <= 0 ) return;
		
		document.get_iter_at_mark (ref iter, document.get_insert ());
		TextIter.set_line (ref iter, linenum - 1);
		document.place_cursor (ref iter);
		scroll_to_iter (ref iter, 0, false, 0, 0);
	}
	
	public void insert_new_line () {
		TextIter iter;
		unichar ichar;
		String indent = new String ("");
		string tmp;
		// save the indentation of the current line
		document.get_iter_at_mark (ref iter, document.get_insert ());
		TextIter.set_line_offset (ref iter, 0);
		while ( TextIter.get_char (ref iter) == '\t' || TextIter.get_char (ref iter) == ' ' ) {
			ichar = TextIter.get_char (ref iter);
			indent.append_unichar (ichar);
			TextIter.forward_char (ref iter);
		}
		
		// move the cursor to the end of the line and insert a newline character with indentation
		if ( !TextIter.ends_line (ref iter) ) TextIter.forward_to_line_end (ref iter);
		document.insert_interactive (ref iter, "\n" + indent.str, -1, true);
		document.get_iter_at_mark (out iter, document.get_insert ());
		if ( !TextIter.ends_line (ref iter) ) {
			TextIter.forward_to_line_end (ref iter);
			TextIter.forward_to_line_end (ref iter);
		}
		document.place_cursor (ref iter);
	}
	
	public void tag_selection () {
		TextIter s, e;
		
		if (document.get_selection_bounds (ref s, ref e)) {
			add_tag (s, e);
		}
	}
	
	
	public void tag_and_find_next () {
		TextIter s, e;
		
		if (document.get_selection_bounds (ref s, ref e)) {
			add_tag (s, e);
			find (document.get_text (ref s, ref e, false), e, true);
		}
	}
	
	public void tag_search_results (string str) {
		TextIter s, e;
		TextIter match, match_end;
		
		document.get_bounds (out s, out e);
			
		while ( SourceIter.forward_search (ref s, str, (SourceSearchFlags)0, out match, out match_end, ref e) ) {
			add_tag (match, match_end);
			s = match_end;
		}
	}
	
	public void remove_tags () {
		TextIter s, e;
		int os, oe, ts, te;
		
		if (!document.get_selection_bounds (ref s, ref e))
			document.get_bounds (ref s, ref e);
			
		os = TextIter.get_offset (ref s);
		oe = TextIter.get_offset (ref e);
		
		SList<Tag> tmp_list;
		foreach (Tag tag in tag_list) {
			document.get_iter_at_mark (out s, tag.begin);
			document.get_iter_at_mark (out e, tag.end);
			ts = TextIter.get_offset (ref s);
			te = TextIter.get_offset (ref e);
			
			if ( ts < os || te > oe ) {
				tmp_list.append (tag);
			}
		}
		tag_list = tmp_list;
		
		update_tags ();
	}
	
	public void replace (string str) {
		TextIter s, e;
		
		document.begin_user_action ();
		foreach (Tag tag in tag_list) {
			document.get_iter_at_mark (out s, tag.begin);
			document.get_iter_at_mark (out e, tag.end);
			document.delete (ref s, ref e);
			document.insert_interactive (ref s, str, -1, true);
			document.get_iter_at_mark (out s, tag.begin);
			document.get_iter_at_mark (out e, tag.end);
			document.apply_tag_by_name ("tag", ref s, ref e);
		}
		document.end_user_action ();
	}
	
	private void add_tag (TextIter s, TextIter e) {
		document.apply_tag_by_name ("tag", ref s, ref e);
		var tag = new Tag ();
		tag.begin = document.create_mark (null, ref s, true);
		tag.end = document.create_mark (null, ref e, false);
		tag_list.append (tag);	
	}
	
	private void update_tags () {
		TextIter s, e;
		
		document.get_bounds (ref s, ref e);
		document.remove_tag_by_name ("tag", ref s, ref e);
		foreach (Tag tag in tag_list) {
			document.get_iter_at_mark (out s, tag.begin);
			document.get_iter_at_mark (out e, tag.end);
			document.apply_tag_by_name ("tag", ref s, ref e);
		}
	}
	
	private void on_font_changed (uint cnxn_id, GConf.Entry entry) {
		stdout.printf ("font changed\n");
	}
	
	private bool on_focus_out (GLib.Object o, Gdk.EventFocus e) {
		return_if_fail (document != null);
		
		// save document state
		save_state ();
		if ( wrap_mode == Gtk.WrapMode.NONE ) {
			document.metadata.@set ("word-wrap", "false");
		} else {
			document.metadata.@set ("word-wrap", "true");
		}
		document.metadata.write ();
		
		document.save ();
		return false;
	}
	
	/* these need to be default handler overrides instead, otherwise things are getting pasted twice... */
	private void on_cut_clipboard (GLib.Object ow) {
		stdout.printf("cut\n");
		//block_cut_clipboard ();
	}
	
	private void on_copy_clipboard (GLib.Object o) {
		stdout.printf("copy\n");
		//block_copy_clipboard ();
	}
	
	private void on_paste_clipboard (GLib.Object o) {
		stdout.printf("past\n");
		//block_paste_clipboard ();
	}
}
